@@ -1,10 +1,10 @@
 package Net::DNS::Resolver;
 
 #
-# $Id: Resolver.pm 1406 2015-10-05 08:25:49Z willem $
+# $Id: Resolver.pm 1446 2016-01-26 10:21:56Z willem $
 #
 use vars qw($VERSION);
-$VERSION = (qw$LastChangedRevision: 1406 $)[1];
+$VERSION = (qw$LastChangedRevision: 1446 $)[1];
 
 =head1 NAME
 
@@ -98,6 +98,10 @@ A space-separated list of domains to put in the search list.
 
 A space-separated list of nameservers to query.
 
+=item options
+
+A space-separated list of key:value items.
+
 =back
 
 Except for F</etc/resolv.conf>, files will only be read if owned by the
@@ -161,8 +165,6 @@ Domain name suffix to be appended to queries of unqualified names.
 
 =item usevc
 
-=item stayopen
-
 =item igntc
 
 =item defnames
@@ -359,21 +361,17 @@ The default is port 53.
 
 =head2 srcport
 
-    print 'sending queries from port ', $resolver->srcport, "\n";
     $resolver->srcport(5353);
 
-Gets or sets the port from which queries are sent.
+Sets the port from which queries are sent.
 The default is 0, meaning any port.
 
 =head2 srcaddr
-
-    print 'sending queries from address ', $resolver->srcaddr, "\n";
     $resolver->srcaddr('192.0.2.1');
 
-Gets or sets the source address from which queries are sent.
+Sets the source address from which queries are sent.
 Convenient for forcing queries from a specific interface on a
-multi-homed host.
-The default is 0, meaning any local address.
+multi-homed host.  The default is to use any local address.
 
 
 =head2 bgsend
@@ -394,58 +392,61 @@ of strings.  The record type and class can be omitted; they default to
 A and IN.  If the name looks like an IP address (IPv4 or IPv6),
 an appropriate PTR query will be performed.
 
-Returns an L<IO::Select> object which serves as a handle which is passed
-to subsequent invocations of the C<bgisready> and C<bgread> methods.
+Returns an opaque handle which is passed to subsequent invocations of
+the C<bgbusy> and C<bgread> methods.
 Errors are indicated by returning C<undef> in which case the
 reason for failure can be found by calling the errorstring method.
 
-The program must determine when the underlying socket is ready for
-reading by calling C<bgisready>, then call C<bgread> to get the response. 
-C<bgread> returns C<undef> if no response is received or timeout occurred. 
+The program may determine when the handle is ready for reading by
+calling C<bgbusy>.
+
+The response C<Net::DNS::Packet> object is obtained by calling C<bgread>.
 
 B<BEWARE>:
-Answers may not fit in an UDP packet and might be truncated. Truncated 
-packets will B<not> be retried over TCP automatically and should be handled
-by the caller.
+Programs should make no assumptions about the nature of the handles
+returned by B<bgsend> which should be used strictly as described here.
 
 
 =head2 bgread
 
     $packet = $resolver->bgread($handle);
 
-    if ($packet->header->tc) { 
-	# Retry over TCP (blocking).
-    }
-
-Reads the answer from a background query (see L</bgsend>).
+Reads the answer from a background query.
 The argument is the handle returned by C<bgsend>.
 
-Returns a C<Net::DNS::Packet> object or C<undef> if an error
-was encountered or the transaction timed out.
+Returns a C<Net::DNS::Packet> object or C<undef> if no response was
+received or timeout occurred. 
 
 
-=head2 bgisready
+=head2 bgbusy
 
     $handle = $resolver->bgsend( 'foo.example.com' );
 
-    until ($resolver->bgisready($handle)) {
-	# do some other processing
+    while ($resolver->bgbusy($handle)) {
+	...
     }
 
     $packet = $resolver->bgread($handle);
 
-    if ($packet->header->tc) { 
-	# Retry over TCP (blocking).
+Returns true while awaiting the response or for the transaction to time out.
+The argument is the handle returned by C<bgsend>.
+
+Truncated UDP packets will be retried over TCP transparently while
+continuing to assert busy to the caller.
+
+
+=head2 bgisready
+
+    until ($resolver->bgisready($handle)) {
+	...
     }
 
-Returns true when the response is available for reading
-or the transaction timed out.
-The argument is the handle returned by C<bgsend>.
+C<bgisready> is the logical complement of C<bgbusy> which is retained
+for backward compatibility.
 
 
 =head2 tsig
 
-    $tsig = $resolver->tsig;
     $resolver->tsig( $tsig );
 
     $resolver->tsig( 'Khmac-sha1.example.+161+24053.private' );
@@ -460,9 +461,9 @@ The argument is the handle returned by C<bgsend>.
 
     $resolver->tsig( undef );
 
-Get or set the TSIG record used to automatically sign outgoing
-queries and updates.  Call with an undefined argument, 0 or ''
-to turn off automatic signing.
+Set the TSIG record used to automatically sign outgoing queries, zone
+transfers and updates.  Call with an undefined argument, 0 or '' to
+turn off automatic signing.
 
 The default resolver behavior is not to sign any packets.  You must
 call this method to set the key if you would like the resolver to
@@ -577,9 +578,11 @@ in which case the default is true.
     $resolver->persistent_udp(1);
 
 Get or set the persistent UDP setting.
-If true, Net::DNS will keep a single UDP socket open for all queries.
-This is useful if you are using UDP and need to make a lot of queries
-or updates.
+If true, a Net::DNS resolver will use the same UDP socket
+for all queries within each address family.
+
+This avoids the cost of creating and tearing down UDP sockets,
+but also defeats source port randomisation.
 
 =head2 igntc
 
@@ -710,11 +713,11 @@ The default domain.
 =head2 RES_OPTIONS
 
     # Bourne Shell
-    RES_OPTIONS="retrans:3 retry:2 debug"
+    RES_OPTIONS="retrans:3 retry:2 inet6"
     export RES_OPTIONS
 
     # C Shell
-    setenv RES_OPTIONS "retrans:3 retry:2 debug"
+    setenv RES_OPTIONS "retrans:3 retry:2 inet6"
 
 A space-separated list of resolver options to set.  Options that
 take values are specified as C<option:value>.
@@ -726,8 +729,8 @@ The Net::DNS::Resolver library will enable IPv6 transport if the
 appropriate libraries (IO::Socket::INET6 or IO::Socket::IP) are
 available and the destination nameserver has an IPv6 address.
 
-The force_v4(), force_v6() and prefer_v6() methods with a non-zero
-argument may be used to configure transport selection.
+The force_v4(), force_v6(),prefer_v4 and prefer_v6() methods with a
+non-zero argument may be used to configure transport selection.
 
 The behaviour of the nameserver() method illustrates the transport
 selection mechanism.  If, for example, IPv6 is not available or IPv4
