@@ -1,10 +1,10 @@
 package Net::DNS::Resolver::Base;
 
 #
-# $Id: Base.pm 1434 2015-12-02 14:25:43Z willem $
+# $Id: Base.pm 1449 2016-02-01 12:27:12Z willem $
 #
 use vars qw($VERSION);
-$VERSION = (qw$LastChangedRevision: 1434 $)[1];
+$VERSION = (qw$LastChangedRevision: 1449 $)[1];
 
 
 use strict;
@@ -31,7 +31,7 @@ use constant PACKETSZ => 512;
 #  inet_pton is not available on WIN32, so we only use the getaddrinfo
 #  call to translate IP addresses to socketaddress
 #
-#  Three configuration flags, force_v4, prefer_v6 and force_v6,
+#  The configuration options force_v4, force_v6, prefer_v4 and prefer_v6
 #  are provided to control IPv6 behaviour for test purposes.
 #
 # Olaf Kolkman, RIPE NCC, December 2003.
@@ -72,9 +72,10 @@ sub _untaint {
 {
 	my $defaults = bless {
 		nameserver4	=> ['127.0.0.1'],
-		nameserver6	=> ( IPv6 ? ['::1'] : [] ),
+		nameserver6	=> ['::1'],
 		port		=> 53,
-		srcaddr		=> 0,
+		srcaddr4	=> '0.0.0.0',
+		srcaddr6	=> '::',
 		srcport		=> 0,
 		searchlist	=> [],
 		retrans		=> 5,
@@ -96,9 +97,9 @@ sub _untaint {
 		adflag		=> 0,	# see RFC6840, 5.7
 		cdflag		=> 0,	# see RFC6840, 5.9
 		udppacketsize	=> 0,	# value bounded below by PACKETSZ
-		force_v4	=> 0,	# only relevant if IPv6 is supported
-		force_v6	=> 0,	#
-		prefer_v6	=> 0,	# prefer v6, otherwise prefer v4
+		force_v4	=> ( IPv6 ? 0 : 1 ),
+		force_v6	=> 0,	# only relevant if IPv6 is supported
+		prefer_v4	=> 1,	# prefer v6 by default
 		},
 			__PACKAGE__;
 
@@ -131,11 +132,10 @@ my %public_attr = map { $_ => $_ } qw(
 		dnssec
 		adflag
 		cdflag
+		force_v4
+		force_v6
 		prefer_v4
 		prefer_v6
-
-		ignqrid
-		stayopen
 		);
 
 
@@ -261,10 +261,8 @@ sub string {
 	my $self = shift;
 	$self = $self->_defaults unless ref($self);
 
-	my $IP6line = "prefer_v6\t= $self->{prefer_v6}\tforce_v6\t= $self->{force_v6}";
-	my $IP6conf = IPv6 ? $IP6line : '(no IPv6 transport)';
-	my @nslist  = $self->nameservers();
-	my $domain  = $self->domain;
+	my @nslist = $self->nameservers();
+	my $domain = $self->domain;
 	return <<END;
 ;; RESOLVER state:
 ;; domain	= $domain
@@ -274,11 +272,10 @@ sub string {
 ;; retrans	= $self->{retrans}	retry		= $self->{retry}
 ;; recurse	= $self->{recurse}	igntc		= $self->{igntc}
 ;; usevc	= $self->{usevc}	port		= $self->{port}
-;; srcaddr	= $self->{srcaddr}	srcport		= $self->{srcport}
 ;; tcp_timeout	= $self->{tcp_timeout}	persistent_tcp	= $self->{persistent_tcp}
 ;; udp_timeout	= $self->{udp_timeout}	persistent_udp	= $self->{persistent_udp}
-;; debug	= $self->{debug}	force_v4	= $self->{force_v4}
-;; $IP6conf
+;; prefer_v4	= $self->{prefer_v4}	force_v4	= $self->{force_v4}
+;; debug	= $self->{debug}	force_v6	= $self->{force_v6}
 END
 
 }
@@ -336,18 +333,18 @@ sub nameservers {
 
 	unless ( defined wantarray ) {
 		$self->{nameserver4} = \@ipv4;
-		$self->{nameserver6} = \@ipv6 if IPv6;
+		$self->{nameserver6} = \@ipv6;
 		return;
 	}
 
 	if ( scalar @_ ) {
 		$self->{nameserver4} = \@ipv4;
-		$self->{nameserver6} = \@ipv6 if IPv6;
+		$self->{nameserver6} = \@ipv6;
 	}
 
 	my @ns4 = $self->force_v6 ? () : @{$self->{nameserver4}};
 	my @ns6 = $self->force_v4 ? () : @{$self->{nameserver6}};
-	my @returnval = $self->prefer_v6 ? ( @ns6, @ns4 ) : ( @ns4, @ns6 );
+	my @returnval = $self->{prefer_v4} ? ( @ns4, @ns6 ) : ( @ns6, @ns4 );
 
 	return @returnval if scalar @returnval;
 
@@ -365,29 +362,18 @@ sub _cname_addr {
 	# TODO 20081217
 	# This code does not follow CNAME chains, it only looks inside the packet.
 	# Out of bailiwick will fail.
-	my @addr;
-	my $packet = shift || return @addr;
+	my @null;
+	my $packet = shift || return @null;
 	my $names = shift;
 
 	map $names->{$_->qname}++, $packet->question;
 	map $names->{$_->cname}++, grep $_->can('cname'), $packet->answer;
 
-	foreach my $rr ( grep $names->{$_->name}, $packet->answer ) {
-		push( @addr, $rr->address ) if $rr->can('address');
-	}
-
-	return @addr;
+	my @addr = grep $_->can('address'), $packet->answer;
+	map $_->address, grep $names->{$_->name}, @addr;
 }
 
 
-# if ($self->{udppacketsize} > PACKETSZ
-# then we use EDNS and $self->{udppacketsize}
-# should be taken as the maximum packet_data length
-sub _packetsz {
-	my $udpsize = shift->{udppacketsize} || 0;
-	return $udpsize > PACKETSZ ? $udpsize : PACKETSZ;
-}
-
 sub answerfrom {
 	my $self = shift;
 	$self->{answerfrom} = shift if scalar @_;
@@ -402,7 +388,6 @@ sub errorstring {
 
 sub _reset_errorstring {
 	my $self = shift;
-
 	$self->errorstring( $self->_defaults->{errorstring} );
 }
 
@@ -412,7 +397,8 @@ sub query {
 	my $name = shift || '.';
 
 	# resolve name containing no dots or colons by appending domain
-	my $fqdn = $self->{defnames} && ( $name !~ /[:.]/ ) ? join( '.', $name, $self->domain ) : $name;
+	my @sfix = $self->{defnames} && $name !~ m/[:.]/ ? $self->domain : ();
+	my $fqdn = join '.', $name, @sfix;
 
 	$self->_diag( 'query(', $fqdn, @_, ')' );
 
@@ -428,22 +414,17 @@ sub search {
 	return $self->query(@_) unless $self->{dnsrch};
 
 	my $name = shift || '.';
+	my @sfix = @{$self->{searchlist}};
+	my @list = $name =~ m/[.]/ ? ( undef, @sfix ) : ( @sfix, undef );
 
-	# resolve name by trying as absolute name, then applying searchlist
-	my @searchlist = ( undef, @{$self->{searchlist}} );
-
-	# attempt to resolve name with no dots or colons by applying suffix
-	my @list = $name =~ m/[:.]/ ? (undef) : @searchlist;
-
-	foreach my $suffix (@list) {
+	foreach my $suffix ( $name =~ m/:|\.\d*$/ ? undef : @list ) {
 		my $fqname = $suffix ? join( '.', $name, $suffix ) : $name;
 
 		$self->_diag( 'search(', $fqname, @_, ')' );
 
 		my $packet = $self->send( $fqname, @_ ) || next;
 
-		next unless $packet->header->rcode eq "NOERROR";
-		return $packet if $packet->header->ancount;	# answer found
+		return $packet->header->ancount ? $packet : next;
 	}
 
 	return undef;
@@ -458,10 +439,10 @@ sub send {
 	return $self->_send_tcp( $packet, $packet_data )
 			if $self->{usevc} || length $packet_data > $self->_packetsz;
 
-	my $ans = $self->_send_udp( $packet, $packet_data );
+	my $ans = $self->_send_udp( $packet, $packet_data ) || return;
 
 	return $ans if $self->{igntc};
-	return $ans unless $ans && $ans->header->tc;
+	return $ans unless $ans->header->tc;
 
 	$self->_diag('packet truncated: retrying using TCP');
 	$self->_send_tcp( $packet, $packet_data );
@@ -477,68 +458,40 @@ sub _send_tcp {
 	my $lastanswer;
 
 	foreach my $ns (@ns) {
-		my $sock = $self->_create_tcp_socket($ns) || next;
+		my $socket = $self->_create_tcp_socket($ns) || next;
 
 		$self->_diag( 'tcp send:', $length, 'bytes' );
 
-		unless ( $sock->send( pack 'n a*', $length, $packet_data ) ) {
-			$self->_diag( $self->errorstring($!) );
-			next;
-		}
+		$socket->send( pack 'n a*', $length, $packet_data );
+		$self->errorstring($!);
 
-		my $sel	    = IO::Select->new($sock);
-		my $timeout = $self->{tcp_timeout};
-		if ( $sel->can_read($timeout) ) {
-			my $buf = _read_tcp( $sock, INT16SZ );
-			next unless length($buf);		# failed to get anything
-			my $len = unpack 'n', $buf;
-			next unless $len;			# cannot determine size
-
-			unless ( $sel->can_read($timeout) ) {
-				$self->_diag( $self->errorstring('timeout') );
-				next;
-			}
-
-			$buf = _read_tcp( $sock, $len );
+		my $sel = IO::Select->new($socket);
+		if ( $sel->can_read( $self->{tcp_timeout} ) ) {
+			my $buffer = _read_tcp($socket);
 			$self->answerfrom($ns);
+			$self->_diag( "answer from [$ns]", length($buffer), 'bytes' );
 
-			my $received = length $buf;
-			$self->_diag( 'received', $received, 'bytes' );
+			my $ans = Net::DNS::Packet->new( \$buffer, $self->{debug} );
+			$self->errorstring($@);
 
-			unless ( $received == $len ) {
-				$self->errorstring("expected $len bytes, received $received");
-				next;
-			}
-
-			my $ans = Net::DNS::Packet->new( \$buf, $self->{debug} );
-
-			unless ( defined $ans ) {
-				$self->errorstring($@);
-			} else {
-				my $rcode = $ans->header->rcode;
-				$self->errorstring( $@ . $rcode );
+			if ($ans) {
+				next unless $ans->verify($packet);
 
 				$ans->answerfrom($ns);
 
-				if ( $rcode ne "NOERROR" && $rcode ne "NXDOMAIN" ) {
-					$self->_diag("RCODE: $rcode; try next nameserver");
-					$lastanswer = $ans;
-					next;
-				}
+				my $rcode = $ans->header->rcode;
+				return $ans if $rcode eq "NOERROR";
+				return $ans if $rcode eq "NXDOMAIN";
+
+				$self->_diag("RCODE: $rcode; try next nameserver");
+				$lastanswer = $ans;
 			}
-			return $ans;
-		} else {
-			$self->errorstring('timeout');
 		}
 	}
 
-	if ($lastanswer) {
-		$self->errorstring( $lastanswer->header->rcode );
-		return $lastanswer;
-	}
-
+	$self->errorstring( $lastanswer->header->rcode ) if $lastanswer;
 	$self->_diag( $self->errorstring );
-	return;
+	return $lastanswer;
 }
 
 
@@ -555,7 +508,7 @@ sub _send_udp {
 
 	foreach my $ip ( $self->nameservers ) {
 		my $socket = $self->_create_udp_socket($ip) || next;
-		my $dst_sockaddr = $self->_create_dst_sockaddr( $ip, $port ) || next;
+		my $dst_sockaddr = $self->_create_dst_sockaddr( $ip, $port );
 		push @ns, [$socket, $ip, $dst_sockaddr];
 	}
 
@@ -583,8 +536,7 @@ NAMESERVER: foreach my $ns (@ns) {
 
 			# handle failure to detect taint inside socket->send()
 			# uncoverable branch true
-			die 'Insecure dependency while running with -T switch'
-					if _tainted($dst_sockaddr);
+			die 'Insecure dependency while running with -T switch' if _tainted($dst_sockaddr);
 
 			$sel->add($ns);
 
@@ -593,38 +545,36 @@ NAMESERVER: foreach my $ns (@ns) {
 				my ( $socket, $ip ) = @$ready;
 				$sel->remove($ready);
 
-				my $buf = '';
-				unless ( $socket->recv( $buf, $self->_packetsz ) ) {
-					$self->_diag( "recv ERROR [$ip]", $ns->[3] = $self->errorstring($!) );
-					next;
-				}
+				my $peer = $socket->peerhost;
+				$self->answerfrom($peer);
 
-				my $peerhost = $socket->peerhost;
-				$self->answerfrom($peerhost);
-				$self->_diag( "answer from [$peerhost]", 'length', length($buf) );
+				my $buffer = _read_udp( $socket, $self->_packetsz );
+				$self->errorstring( $ns->[3] = $! ) unless $buffer;
+				$self->_diag( "answer from [$peer]", length($buffer), 'bytes' );
 
-				my $ans = Net::DNS::Packet->new( \$buf, $self->{debug} );
+				my $ans = Net::DNS::Packet->new( \$buffer, $self->{debug} );
+				$self->errorstring($@);
 
-				unless ( defined $ans ) {
-					$self->errorstring($@);
-				} else {
+				if ($ans) {
 					my $header = $ans->header;
-					my $rcode  = $header->rcode;
-					$self->errorstring( $@ . $rcode );
-
-					$ans->answerfrom($peerhost);
-
 					next unless $header->qr;
 					next unless $header->id == $packet->header->id;
 
-					if ( $rcode ne "NOERROR" && $rcode ne "NXDOMAIN" ) {
-						my $msg = $ns->[3] = "RCODE: $rcode";
-						$self->_diag("$msg; try next nameserver");
-						$lastanswer = $ans;
-						next NAMESERVER;
+					unless ( $ans->verify($packet) ) {
+						$self->errorstring( $ns->[3] = $ans->verifyerr );
+						next;
 					}
+
+					$ans->answerfrom($peer);
+
+					my $rcode = $header->rcode;
+					return $ans if $rcode eq "NOERROR";
+					return $ans if $rcode eq "NXDOMAIN";
+
+					my $msg = $ns->[3] = "RCODE: $rcode";
+					$self->_diag("$msg; try next nameserver");
+					$lastanswer = $ans;
 				}
-				return $ans;
 			}					#SELECTOR LOOP
 		}						#NAMESERVER LOOP
 		no integer;
@@ -639,12 +589,14 @@ NAMESERVER: foreach my $ns (@ns) {
 
 
 sub bgsend {
-	my $self = shift;
+	my $self	= shift;
+	my $packet	= $self->_make_query_packet(@_);
+	my $packet_data = $packet->data;
 
-	my $packet = $self->_make_query_packet(@_);
+	return $self->_bgsend_tcp( $packet, $packet_data )
+			if $self->{usevc} || length $packet_data > $self->_packetsz;
 
-	return $self->_bgsend_tcp( $packet, $packet->data ) if $self->{usevc};
-	return $self->_bgsend_udp( $packet, $packet->data );
+	return $self->_bgsend_udp( $packet, $packet_data );
 }
 
 
@@ -655,21 +607,19 @@ sub _bgsend_tcp {
 
 	my $port = $self->{port};
 
-	foreach my $ns ( $self->nameservers ) {
-		my $socket = $self->_create_tcp_socket($ns) || next;
+	foreach my $ip ( $self->nameservers ) {
+		my $socket = $self->_create_tcp_socket($ip) || next;
 
-		$self->_diag( 'bgsend', "[$ns]:$port" );
+		$self->_diag( 'bgsend', "[$ip]:$port" );
 
 		my $length = length $packet_data;
 		my $tcp_packet = pack 'n a*', $length, $packet_data;
 
-		unless ( $socket->send($tcp_packet) ) {
-			$self->errorstring("send: [$ns]:$port  $!");
-			next;
-		}
+		$socket->send($tcp_packet);
+		$self->errorstring($!);
 
-		my $expire = time() + $self->{tcp_timeout} || 0;
-		${*$socket}{net_dns_bg} = [$expire, $ns, $packet->header->id];
+		my $expire = time() + $self->{tcp_timeout};
+		${*$socket}{net_dns_bg} = [$expire, 0, $packet->header->id, $ip];
 		return $socket;
 	}
 
@@ -685,24 +635,22 @@ sub _bgsend_udp {
 
 	my $port = $self->{port};
 
-	foreach my $ns ( $self->nameservers ) {
-		my $socket = $self->_create_udp_socket($ns) || next;
+	foreach my $ip ( $self->nameservers ) {
+		my $socket = $self->_create_udp_socket($ip) || next;
+		my $dst_sockaddr = $self->_create_dst_sockaddr( $ip, $port );
 
-		my $dst_sockaddr = $self->_create_dst_sockaddr( $ns, $port ) || next;
+		$self->_diag( 'bgsend', "[$ip]:$port" );
 
-		$self->_diag( 'bgsend', "[$ns]:$port" );
-
-		unless ( $socket->send( $packet_data, 0, $dst_sockaddr ) ) {
-			$self->errorstring("send: [$ns]:$port  $!");
-			next;
-		}
+		my $ok = $socket->send( $packet_data, 0, $dst_sockaddr );
+		$self->errorstring($!);
+		next unless $ok;
 
 		# handle failure to detect taint inside $socket->send()
 		# uncoverable branch true
 		die 'Insecure dependency while running with -T switch' if _tainted($dst_sockaddr);
 
-		my $expire = time() + $self->{udp_timeout} || 0;
-		${*$socket}{net_dns_bg} = [$expire, $ns, $packet->header->id];
+		my $expire = time() + $self->{udp_timeout};
+		${*$socket}{net_dns_bg} = [$expire, 1, $packet->header->id, $ip];
 		return $socket;
 	}
 
@@ -711,66 +659,80 @@ sub _bgsend_udp {
 }
 
 
-sub bgisready {
+sub bgbusy {
 	my $self = shift;
-	my $sock = shift || return 1;
+	my ($sock) = @_;
+	return unless $sock;
 
-	return scalar( IO::Select->new($sock)->can_read(0.0) ) || do {
-		return 0 unless $self->{udp_timeout};
-		my $appendix = ${*$sock}{net_dns_bg} || [];
-		my ($expire) = @$appendix;
-		time() > ( $expire || return 1 );
-	};
+	my $appendix = ${*$sock}{net_dns_bg} || [];
+	my ( $expire, $udp ) = ( @$appendix, 0, 1 );
+	return if ref($udp);
+
+	unless ( IO::Select->new($sock)->can_read(0) ) {
+		my $time = time();
+		my ($expire) = ( @$appendix, $time );
+		return $time <= $expire;
+	}
+
+	return unless $udp;
+
+	my $ans = $self->bgread($sock);
+	return unless $ans;
+
+	$$appendix[1] = $ans;
+	return unless $ans->header->tc;
+	return if $self->{igntc};
+
+	$self->_diag('packet truncated: retrying using TCP');
+	my $packet = new Net::DNS::Packet();
+	$packet->{question} = $ans->{question};
+	my $tcp = $self->_bgsend_tcp( $packet, $packet->data );
+	$_[0] = $tcp if $tcp;
+	return defined $tcp;
+}
+
+
+sub bgisready {				## historical
+	!&bgbusy;						# uncoverable pod
 }
 
 
 sub bgread {
 	my $self = shift;
-	my $sock = shift || return undef;
+	my $sock = shift || return;
+
+	my $appendix = ${*$sock}{net_dns_bg} || [];
+	my ( $x, $udp ) = ( @$appendix, 0, 1 );
+
+	return $udp if ref($udp);
 
 	my $time = time();
-	my $appendix = ${*$sock}{net_dns_bg} || [$time];
-	my ( $expire, $ip, $id ) = @$appendix;
+	my ( $expire, $u, $qid, $ip ) = ( @$appendix, $time, 1, 0, '' );
 
 	my $select  = IO::Select->new($sock);
 	my $timeout = $expire - $time;
 	return undef unless $select->can_read( $timeout > 0 ? $timeout : 0 );
 
 	my $buffer;
-	if ( $self->{usevc} ) {
-		$buffer = _read_tcp( $sock, INT16SZ );
-		return undef unless length($buffer);		# failed to get length
-		my $len = unpack 'n', $buffer;
-		return undef unless $len;			# can not determine size
-
-		return undef unless $select->can_read( $timeout > 0 ? $timeout : 0 );
-
-		$buffer = _read_tcp( $sock, $len );
-		$self->answerfrom($ip);				# $sock->peerhost unreliable
-		$self->_diag("answer from [$ip]");
-
-	} else {
-		unless ( $sock->recv( $buffer, $self->_packetsz ) ) {
-			$self->errorstring($!);
-			return undef;
-		}
-
+	if ($udp) {
 		my $peerhost = $sock->peerhost;
 		$self->answerfrom($peerhost);
-		$self->_diag("answer from [$peerhost]");
+		$buffer = _read_udp( $sock, $self->_packetsz );
+		$self->_diag( "answer from [$peerhost]", length($buffer), 'bytes' );
+
+	} else {
+		$self->answerfrom($ip);				# $sock->peerhost unreliable
+		$buffer = _read_tcp($sock);
+		$self->_diag( "answer from [$ip]", length($buffer), 'bytes' );
 	}
 
-
 	my $ans = Net::DNS::Packet->new( \$buffer, $self->{debug} );
+	$self->errorstring($@);
 
-	unless ( defined $ans ) {
-		$self->errorstring($@);
-	} else {
+	if ($ans) {
 		my $header = $ans->header;
-		$self->errorstring( $@ . $header->rcode );
-
 		return undef unless $header->qr;
-		return undef unless not defined($id) or $header->id == $id;
+		return undef unless $header->id == $qid;
 
 		$ans->answerfrom( $self->answerfrom );
 	}
@@ -778,128 +740,78 @@ sub bgread {
 }
 
 
-sub _make_query_packet {
-	my $self = shift;
-	my $packet;
-
-	if ( ref( $_[0] ) and $_[0]->isa('Net::DNS::Packet') ) {
-		$packet = shift;
-	} else {
-		$packet = Net::DNS::Packet->new(@_);
-	}
-
-	my $header = $packet->header;
-
-	$header->rd( $self->{recurse} ) if $header->opcode eq 'QUERY';
-
-	$header->ad(1) if $self->{adflag};			# RFC6840, 5.7
-	$header->cd(1) if $self->{cdflag};			# RFC6840, 5.9
-
-	if ( $self->dnssec ) {
-		$self->_diag('Set EDNS DO flag');
-		$header->ad(0);
-		$header->do(1);
-	}
-
-	$packet->edns->size( $self->{udppacketsize} );		# advertise payload size for local stack
-
-	if ( $self->{tsig_rr} && !grep $_->type eq 'TSIG', $packet->additional ) {
-		$packet->sign_tsig( $self->{tsig_rr} );
-	}
-
-	return $packet;
-}
-
-
 sub axfr {				## zone transfer
 	my $self = shift;
 
-	my $query = $self->_axfr_start(@_);
-	my $verfy = $query ? $query->sigrr : undef;
-	my @rr;
-	my $soa;
+	my ( $verify, @rr, $soa ) = $self->_axfr_start(@_);	# iterator state
 
 	my $iterator = sub {		## iterate over RRs
 		my $rr = shift(@rr);
-		return $rr if scalar @rr;
 
-		if ( ( ref($rr) eq 'Net::DNS::RR::SOA' ) && ( $rr ne $soa ) ) {
-			croak 'improperly terminated AXFR' unless $rr->encode eq $soa->encode;
+		if ( ref($rr) eq 'Net::DNS::RR::SOA' ) {
+			return $soa = $rr unless $soa;
+			croak 'improperly terminated AXFR' if $rr->encode ne $soa->encode;
 			return $self->{axfr_sel} = undef;
 		}
 
-		my $reply = $self->_axfr_next() || return $rr;	# end of packet
-		$verfy = $reply->verify($verfy) || croak $reply->verifyerr if $verfy;
-		$self->_diag( $verfy ? 'verified' : 'not verified' );
+		return $rr if scalar @rr;
+
+		my $reply;
+		( $reply, $verify ) = $self->_axfr_next($verify);
+		return $self->{axfr_sel} = undef unless $reply;
 		@rr = $reply->answer;
 		return $rr;
 	};
 
 	$iterator->();						# read initial packet
-	($soa) = @rr;
 
 	return $iterator unless wantarray;
 
-	my @zone = shift @rr;					# initial SOA
-	my @null;
-
-	until ( scalar(@rr) && $rr[$#rr]->type eq 'SOA' ) {
-		push @zone, @rr;				# unpack non-terminal packet
-		@rr = @null;
-		my $reply = $self->_axfr_next() || last;
-		$verfy = $reply->verify($verfy) || croak $reply->verifyerr if $verfy;
-		$self->_diag( $verfy ? 'verified' : 'not verified' );
-		@rr = $reply->answer;
+	my @zone;						# assemble whole zone
+	while ( my $rr = $iterator->() ) {
+		push @zone, $rr, @rr;				# copy RRs en bloc
+		@rr = pop(@zone);				# leave last one in @rr
 	}
-
-	my $last = pop @rr;					# unpack final packet
-	push @zone, @rr;
-	$self->{axfr_sel} = undef;
-	return @zone if $last && $last->encode eq $soa->encode;
-	croak 'improperly terminated AXFR';
+	return @zone;
 }
 
 
 sub axfr_start {			## historical
 	my $self = shift;					# uncoverable pod
 	my $iter = $self->{axfr_iter} = $self->axfr(@_);
-	return defined($iter);
+	defined($iter);
 }
 
 
 sub axfr_next {				## historical
-	my $self = shift;					# uncoverable pod
-	my $iter = $self->{axfr_iter} || return undef;
-	$iter->() || return $self->{axfr_iter} = undef;
+	shift->{axfr_iter}->();					# uncoverable pod
 }
 
 
 sub _axfr_start {
 	my $self  = shift;
-	my $dname = shift || $self->domain;
+	my $dname = scalar(@_) ? shift : $self->domain;
 	my @class = @_;
 
-	my $packet = $self->_make_query_packet( $dname, 'AXFR', @class );
+	my $request = $self->_make_query_packet( $dname, 'AXFR', @class );
 
 	$self->_diag("axfr_start( $dname, @class )");
 
 	foreach my $ns ( $self->nameservers ) {
-		my $sock = $self->_create_tcp_socket($ns) || next;
+		my $socket = $self->_create_tcp_socket($ns) || next;
 
 		$self->_diag("axfr_start nameserver [$ns]");
 
-		my $packet_data = $packet->data;
+		my $packet_data = $request->data;
 		my $TCP_msg = pack 'n a*', length($packet_data), $packet_data;
 
-		unless ( $sock->send($TCP_msg) ) {
-			$self->errorstring($!);
-			next;
-		}
+		$socket->send($TCP_msg);
+		$self->errorstring($!);
 
 		$self->{axfr_ns}  = $ns;
-		$self->{axfr_sel} = IO::Select->new($sock);
+		$self->{axfr_sel} = IO::Select->new($socket);
 
-		return $packet;
+		return $request->sigrr ? $request : undef;
 	}
 
 	$self->_diag( $self->errorstring );
@@ -908,113 +820,72 @@ sub _axfr_start {
 
 
 sub _axfr_next {
-	my $self = shift;
+	my ( $self, $verify ) = @_;
 
-	#--------------------------------------------------------------
-	# Read the length of the response packet.
-	#--------------------------------------------------------------
-
-	my $select  = $self->{axfr_sel} || return;
-	my $timeout = $self->{tcp_timeout};
-	my @ready   = $select->can_read($timeout);
-	unless (@ready) {
-		$self->errorstring('timeout');
-		return;
-	}
-
-	my $buf = _read_tcp( $ready[0], INT16SZ );
-	unless ( length $buf ) {
-		$self->errorstring('truncated zone transfer');
-		return;
-	}
-
-	my ($len) = unpack( 'n', $buf );
-	unless ($len) {
-		$self->errorstring('truncated zone transfer');
-		return;
-	}
+	my $select = $self->{axfr_sel} || return;
+	my ($sock) = $select->can_read( $self->{tcp_timeout} );
+	croak 'improperly terminated AXFR' unless $sock;
 
 	#--------------------------------------------------------------
 	# Read the response packet.
 	#--------------------------------------------------------------
 
-	@ready = $select->can_read($timeout);
-	unless (@ready) {
-		$self->errorstring('timeout');
-		return;
-	}
+	my $buffer = _read_tcp($sock);
+	$self->_diag( 'received', length($buffer), 'bytes' );
 
-	$buf = _read_tcp( $ready[0], $len );
-
-	my $received = length $buf;
-	$self->_diag( 'received', $received, 'bytes' );
-
-	unless ( $received == $len ) {
-		$self->_diag( $self->errorstring("expected $len bytes, received $received") );
-
-	} elsif ( my $ans = Net::DNS::Packet->new( \$buf, $self->{debug} ) ) {
-		my $rcode = $ans->header->rcode;
-		$ans->answerfrom( $self->{axfr_ns} );
-		$self->_diag( $self->errorstring("RCODE from server: $rcode") );
-		return $ans if $rcode eq 'NOERROR';
-
-	} else {
-		$self->_diag( $self->errorstring("no packet: $@") );
-	}
-	return;
-}
-
-
-sub tsig {
-	my $self = shift;
-
-	return $self->{tsig_rr} unless scalar @_;
-	$self->{tsig_rr} = eval {
-		local $SIG{__DIE__};
-		require Net::DNS::RR::TSIG;
-		Net::DNS::RR::TSIG->create(@_);
-	};
-	croak "$@ unable to create TSIG record" if $@;
-}
-
-
-sub dnssec {
-	my $self = shift;
-
-	return $self->{dnssec} unless scalar @_;
-
-	# increase default udppacket size if flag set
-	$self->udppacketsize(2048) if $self->{dnssec} = shift;
-
-	return $self->{dnssec};
+	my $packet = Net::DNS::Packet->new( \$buffer );
+	return unless $packet;
+	$packet->answerfrom( $self->{axfr_ns} );
+	my $rcode = $packet->header->rcode;
+	$self->_diag( $self->errorstring("RCODE from server: $rcode") );
+	return $packet unless $verify;
+	return ( $packet, $verify ) if $verify = $packet->verify($verify);
+	croak $packet->verifyerr;
 }
 
 
 #
-# Usage:  $data = _read_tcp($socket, $nbytes);
+# Usage:  $data = _read_tcp($socket);
 #
 sub _read_tcp {
-	my ( $sock, $nbytes ) = @_;
-	my $buf = '';
+	my $socket = shift;
 
-	while ( ( my $unread = $nbytes - length $buf ) > 0 ) {
+	my $size_buf = '';
+	$socket->recv( $size_buf, INT16SZ );
+	my ($unread) = unpack 'n*', $size_buf;
+
+	my $buffer = '';
+	while ($unread) {
 
 		# During some of my tests recv() returned undef even
 		# though there wasn't an error.	 Checking for the amount
 		# of data read appears to work around that problem.
 
 		my $read_buf = '';
-		$sock->recv( $read_buf, $unread ) || do {
-			unless ( length($read_buf) ) {
-				warn "ERROR: tcp recv failed: $!\n";
-				last;
-			}
-		};
+		$socket->recv( $read_buf, $unread );
 
-		$buf .= $read_buf;
+		my $read = length $read_buf;
+		last unless $read;
+
+		$buffer .= $read_buf;
+		$unread -= $read;
 	}
 
-	return $buf;
+	warn "ERROR: tcp recv failed: $!\n" if $unread;
+	return $buffer;
+}
+
+
+#
+# Usage:  $data = _read_udp($socket, $length);
+#
+sub _read_udp {
+	my $socket = shift;
+	my $length = shift;
+
+	my $buffer = '';
+	warn "ERROR: udp recv failed: $!\n" unless $socket->recv( $buffer, $length );
+	return $buffer;
 }
 
 
@@ -1024,24 +895,22 @@ sub _create_tcp_socket {
 
 	my $dstport  = $self->{port};
 	my $sock_key = "[$ns]:$dstport";
-	my $sock;
+	my $socket;
 
-	if ( $self->{persistent_tcp} ) {
-		$sock = $self->{persistent}{$sock_key};
-		$self->_diag( 'using persistent socket', $sock_key ) if $sock;
-		return $sock if $sock && $sock->connected;
+	if ( $socket = $self->{persistent}{$sock_key} ) {
+		$self->_diag( 'using persistent socket', $sock_key );
+		return $socket if $socket->connected;
 		$self->_diag('socket disconnected (trying to connect)');
 	}
 
-	my $srcaddr = $self->{srcaddr};
 	my $srcport = $self->{srcport};
 	my $timeout = $self->{tcp_timeout};
 
 	if ( IPv6 && _ip_is_ipv6($ns) ) {
-		my $localaddr = $srcaddr =~ /[:].*[:]/ ? $srcaddr : '::';
+		my $localaddr = $self->{srcaddr6};
 
-		$sock = IO::Socket::IP->new(
-			LocalAddr => $localaddr,
+		$socket = IO::Socket::IP->new(
+			LocalAddr => $self->{srcaddr6},
 			LocalPort => $srcport,
 			PeerAddr  => $ns,
 			PeerPort  => $dstport,
@@ -1050,7 +919,7 @@ sub _create_tcp_socket {
 			)
 				if USE_SOCKET_IP;
 
-		$sock = IO::Socket::INET6->new(
+		$socket = IO::Socket::INET6->new(
 			LocalAddr => $localaddr,
 			LocalPort => ( $srcport || undef ),
 			PeerAddr  => $ns,
@@ -1060,9 +929,9 @@ sub _create_tcp_socket {
 			)
 				unless USE_SOCKET_IP;
 	} else {
-		my $localaddr = $srcaddr =~ /[.].*[.]/ ? $srcaddr : '0.0.0.0';
+		my $localaddr = $self->{srcaddr4};
 
-		$sock = IO::Socket::IP->new(
+		$socket = IO::Socket::IP->new(
 			LocalAddr => $localaddr,
 			LocalPort => $srcport,
 			PeerAddr  => $ns,
@@ -1072,7 +941,7 @@ sub _create_tcp_socket {
 			)
 				if USE_SOCKET_IP;
 
-		$sock = IO::Socket::INET->new(
+		$socket = IO::Socket::INET->new(
 			LocalAddr => $localaddr,
 			LocalPort => ( $srcport || undef ),
 			PeerAddr  => $ns,
@@ -1083,13 +952,9 @@ sub _create_tcp_socket {
 				unless USE_SOCKET_IP;
 	}
 
-	unless ($sock) {
-		$self->_diag( $self->errorstring("connection failed [$ns]") );
-	} elsif ( $self->{persistent_tcp} ) {
-		$self->{persistent}{$sock_key} = $sock;
-	}
-
-	return $sock;
+	$self->{persistent}{$sock_key} = $self->{persistent_tcp} ? $socket : undef;
+	$self->_diag( $self->errorstring("connection failed $sock_key") ) unless $socket;
+	return $socket;
 }
 
 
@@ -1097,18 +962,17 @@ sub _create_udp_socket {
 	my $self = shift;
 	my $ns	 = shift;
 
-	my $srcaddr = $self->{srcaddr};
 	my $srcport = $self->{srcport};
 	my $sock_key;
-	my $sock;
+	my $socket;
 
 	if ( IPv6 && _ip_is_ipv6($ns) ) {
 		$sock_key = 'UDP/IPv6';
-		return $sock if $sock = $self->{persistent}{$sock_key};
+		return $socket if $socket = $self->{persistent}{$sock_key};
 
-		my $localaddr = $srcaddr =~ /[:].*[:]/ ? $srcaddr : '::';
+		my $localaddr = $self->{srcaddr6};
 
-		$sock = IO::Socket::IP->new(
+		$socket = IO::Socket::IP->new(
 			LocalAddr => $localaddr,
 			LocalPort => $srcport,
 			Proto	  => 'udp',
@@ -1116,7 +980,7 @@ sub _create_udp_socket {
 			)
 				if USE_SOCKET_IP;
 
-		$sock = IO::Socket::INET6->new(
+		$socket = IO::Socket::INET6->new(
 			LocalAddr => $localaddr,
 			LocalPort => ( $srcport || undef ),
 			Proto	  => 'udp',
@@ -1125,11 +989,11 @@ sub _create_udp_socket {
 				unless USE_SOCKET_IP;
 	} else {
 		$sock_key = 'UDP/IPv4';
-		return $sock if $sock = $self->{persistent}{$sock_key};
+		return $socket if $socket = $self->{persistent}{$sock_key};
 
-		my $localaddr = $srcaddr =~ /[.].*[.]/ ? $srcaddr : '0.0.0.0';
+		my $localaddr = $self->{srcaddr4};
 
-		$sock = IO::Socket::IP->new(
+		$socket = IO::Socket::IP->new(
 			LocalAddr => $localaddr,
 			LocalPort => $srcport,
 			Proto	  => 'udp',
@@ -1137,7 +1001,7 @@ sub _create_udp_socket {
 			)
 				if USE_SOCKET_IP;
 
-		$sock = IO::Socket::INET->new(
+		$socket = IO::Socket::INET->new(
 			LocalAddr => $localaddr,
 			LocalPort => ( $srcport || undef ),
 			Proto	  => 'udp',
@@ -1146,9 +1010,9 @@ sub _create_udp_socket {
 				unless USE_SOCKET_IP;
 	}
 
-	$self->_diag( $self->errorstring( 'could not get', $sock_key, 'socket' ) ) unless $sock;
-	return $self->{persistent}{$sock_key} = $sock if $self->{persistent_udp};
-	return $sock;
+	$self->{persistent}{$sock_key} = $self->{persistent_udp} ? $socket : undef;
+	$self->_diag( $self->errorstring("could not get $sock_key socket") ) unless $socket;
+	return $socket;
 }
 
 
@@ -1163,13 +1027,11 @@ sub _create_dst_sockaddr {		## create UDP destination sockaddr structure
 		local $^W = 0;					# circumvent perl -w warnings
 
 		my @res = Socket6::getaddrinfo( $ip, $port, AF_INET6, SOCK_DGRAM, 0, AI_NUMERICHOST );
-		if ( scalar(@res) < 5 ) {
-			my ($error) = @res;
-			$self->errorstring("send: $ip\t$error");
-			return;
-		}
+		return $res[3] unless scalar(@res) < 5;
 
-		return $res[3];
+		my ($error) = @res;
+		$self->errorstring("send: $ip\t$error");
+		return;
 
 	} elsif (USE_SOCKET_IP) {
 		no strict;
@@ -1196,6 +1058,46 @@ sub _ip_is_ipv6 {
 }
 
 
+sub _make_query_packet {
+	my $self = shift;
+
+	my ($packet) = @_;
+	if ( ref($packet) ) {
+		my $header = $packet->header;
+		$header->rd( $self->{recurse} ) if $header->opcode eq 'QUERY';
+
+	} else {
+		$packet = Net::DNS::Packet->new(@_);
+
+		my $header = $packet->header;
+		$header->ad( $self->{adflag} );			# RFC6840, 5.7
+		$header->cd( $self->{cdflag} );			# RFC6840, 5.9
+		$header->do(1) if $self->dnssec;
+		$header->rd( $self->{recurse} );
+	}
+
+	$packet->edns->size( $self->{udppacketsize} );		# advertise UDPsize for local stack
+
+	if ( $self->{tsig_rr} ) {
+		$packet->sign_tsig( $self->{tsig_rr} ) unless $packet->sigrr;
+	}
+
+	return $packet;
+}
+
+
+sub dnssec {
+	my $self = shift;
+
+	return $self->{dnssec} unless scalar @_;
+
+	# increase default udppacket size if flag set
+	$self->udppacketsize(2048) if $self->{dnssec} = shift;
+
+	return $self->{dnssec};
+}
+
+
 sub force_v4 {
 	my $self = shift;
 	return $self->{force_v4} unless scalar @_;
@@ -1214,16 +1116,34 @@ sub force_v6 {
 
 sub prefer_v4 {
 	my $self = shift;
-	return $self->{prefer_v6} ? 0 : 1 unless scalar @_;
-	my $value = shift;
-	$self->{prefer_v6} = $value ? 0 : 1;
-	return $value;
+	return $self->{prefer_v4} unless scalar @_;
+	$self->{prefer_v4} = shift() ? 1 : 0;
 }
 
 sub prefer_v6 {
 	my $self = shift;
-	return $self->{prefer_v6} unless scalar @_;
-	$self->{prefer_v6} = shift() ? 1 : 0;
+	$self->{prefer_v4} = shift() ? 0 : 1 if scalar @_;
+	return $self->{prefer_v4} ? 0 : 1;
+}
+
+
+sub srcaddr {
+	my $self = shift;
+	for (@_) {
+		return $self->{srcaddr4} = $_ if _ip_is_ipv4($_);
+		return $self->{srcaddr6} = $_ if _ip_is_ipv6($_);
+	}
+}
+
+
+sub tsig {
+	my $self = shift;
+	$self->{tsig_rr} = eval {
+		local $SIG{__DIE__};
+		require Net::DNS::RR::TSIG;
+		Net::DNS::RR::TSIG->create(@_);
+	};
+	croak "$@ unable to create TSIG record" if $@;
 }
 
 
@@ -1233,6 +1153,14 @@ sub udppacketsize {
 	return $self->_packetsz;
 }
 
+# if ($self->{udppacketsize} > PACKETSZ
+# then we use EDNS and $self->{udppacketsize}
+# should be taken as the maximum packet_data length
+sub _packetsz {
+	my $udpsize = shift->{udppacketsize} || 0;
+	return $udpsize > PACKETSZ ? $udpsize : PACKETSZ;
+}
+
 
 #
 # Keep this method around. Folk depend on it although it is neither documented nor exported.
@@ -1300,9 +1228,11 @@ for all your resolving needs.
 
 =head2 new, domain, searchlist, nameservers, print, string, errorstring,
 
-=head2 search, query, send, bgsend, bgisready, bgread, axfr, answerfrom,
+=head2 search, query, send, bgsend, bgbusy, bgread, axfr, answerfrom,
 
-=head2 force_v4, force_v6, prefer_v4, prefer_v6, udppacketsize, dnssec, tsig
+=head2 force_v4, force_v6, prefer_v4, prefer_v6,
+
+=head2 dnssec, srcaddr, tsig, udppacketsize
 
 See L<Net::DNS::Resolver>.
 
